/*
 * arch/arm/mach-ns9xxx/include/mach/vmalloc.h
 *
 * Copyright (C) 2006 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */
#ifndef __ASM_ARCH_VMALLOC_H
#define __ASM_ARCH_VMALLOC_H

#define VMALLOC_END     (0xf0000000)

#endif /* ifndef __ASM_ARCH_VMALLOC_H */
