/*
 * arch/arm/mach-ns9xxx/generic.c
 *
 * Copyright (C) 2006,2007 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/memory.h>

#include "generic.h"

void __init ns9xxx_init_machine(void)
{
}
