#ifndef __EXCITE_XICAP_H__
#define __EXCITE_XICAP_H__


/* Resource names */
#define XICAP_RESOURCE_FIFO_RX		"xicap_fifo_rx"
#define XICAP_RESOURCE_FIFO_TX		"xicap_fifo_tx"
#define XICAP_RESOURCE_XDMA 		"xicap_xdma"
#define XICAP_RESOURCE_DMADESC		"xicap_dmadesc"
#define XICAP_RESOURCE_PKTPROC  	"xicap_pktproc"
#define XICAP_RESOURCE_IRQ		"xicap_irq"
#define XICAP_RESOURCE_GPI_SLICE	"xicap_gpi_slice"
#define XICAP_RESOURCE_FIFO_BLK		"xicap_fifo_blocks"
#define XICAP_RESOURCE_PKT_STREAM	"xicap_pkt_stream"

#endif /* __EXCITE_XICAP_H__ */
